import math as math
import numpy as np
import pandas as pd
from functools import reduce
from scipy.sparse import csr_matrix

import fds.sdk.OpenRisk
from fds.sdk.OpenRisk.api import operations_api
from fds.sdk.OpenRisk.models import *


class OpenRiskDelta:
    def __init__(self, fds_username: str, fds_api_key: str):
        self.fds_username = fds_username
        self.fds_api_key = fds_api_key

    def calculate_open_risk_holdings(self, date, ptf_ids, ptf_weight, bench_ids, bench_weight, currency, calendar, risk_model):
        configuration = fds.sdk.OpenRisk.Configuration(username=self.fds_username, password=self.fds_api_key)
        # Raw Stats to be requested
        raw_stats = Stats([Stat(name="PortfolioWeights",
                                level=StatCalculationLevel("Security")),
                           Stat(name="BenchmarkWeights",
                                level=StatCalculationLevel("Security")),
                           Stat(name="RawFactorExposure",
                                level=StatCalculationLevel("FactorSecurity"),
                                settings=StatCalculationSettings(
                                    sparse=True)),
                           Stat(name="RawFactorRisk",
                                level=StatCalculationLevel("Factor")),
                           Stat(name="RawFactorCorrelationMatrix",
                                level=StatCalculationLevel("FactorFactor"),
                                settings=StatCalculationSettings(
                                    sparse=True)),
                           Stat(name="RawAssetCovarianceMatrix",
                                level=StatCalculationLevel("SecuritySecurity"),
                                settings=StatCalculationSettings(
                                    correlated_specific_risk=True,
                                    sparse=True))])
        # SDK Request to Open:Risk API
        with fds.sdk.OpenRisk.ApiClient(configuration) as api_client:
            api_instance = operations_api.OperationsApi(api_client)
            version = "v1"
            calculate_from_holdings_request_body = CalculateFromHoldingsRequestBody(
                data=CalculateFromHoldingsRequestData(
                    date=Date(date),
                    holdings=Holdings(
                        portfolio=Holding(ids=ptf_ids, market_values=ptf_weight),
                        benchmark=Holding(ids=bench_ids, market_values=bench_weight)),
                    currency=CurrencyISOCode(currency),
                    calendar=CalendarCode(calendar),
                    risk_model=RiskModelCode(risk_model),
                    stats=raw_stats))
        try:
            return api_instance.calculate_from_holdings(version, calculate_from_holdings_request_body)
        except fds.sdk.OpenRisk.exceptions as e:
            print("Failed to calculate from holdings.")
            raise(e)

    def get_security_df(self, security_ids, universe_date):
        return pd.DataFrame(
            {
                'symbol': security_ids,
                'date': [universe_date] * len(security_ids)
            }
        )

    def get_portfolio_df(self, universe_date):
        return pd.DataFrame(
            {
                'portfolio': ['portfolio'],
                'date': [universe_date]
            }
        )

    def get_factor_df(self, factor_ids, universe_date):
        return pd.DataFrame(
            {
                'factor': factor_ids,
                'date': [universe_date] * len(factor_ids)
            }
        )

    def get_factor_groups_df(self, group_names, universe_date, statlevel_stat, statlevel_value):
        return pd.DataFrame(
            {
                'factorGroups': group_names,
                'date': [universe_date] * len(group_names),
                statlevel_stat: statlevel_value
            }
        )

    def get_factor_security_df(self, statlevel_value):
        if isinstance(statlevel_value, dict):
            return pd.DataFrame(
                csr_matrix(
                    (statlevel_value['value'],
                    statlevel_value['indices'],
                    statlevel_value['indexPointer']),
                    shape=(statlevel_value['rows'],
                    statlevel_value['columns'])
                ).toarray()
            )

        if isinstance(statlevel_value, list):
            return pd.DataFrame(statlevel_value)
        return None

    def convert_json_output_to_df(self, json_openrisk_output):
        """Converts Open Risk API JSON output into DataFrame for a specific date"""
        df_portfolio = None
        df_security = None
        df_factor = None
        df_factorgroup = None
        df_factorfactor = None
        df_factorsecurity = None
        df_securitysecurity = None
        for y, statlevel in enumerate(json_openrisk_output['data']['value']):
            try:
                statlevel_level = statlevel['level']
                statlevel_stat = statlevel['stat']
                statlevel_value = statlevel['value']
            except KeyError as e:
                print("Failed to find a key from the OpenRisk API call output.")
                raise e

            security_ids = None
            try:
                security_ids = json_openrisk_output['meta']['labels']['security']['ids']
            except KeyError:
                pass

            universe_date = None
            try:
                universe_date = json_openrisk_output["meta"]["resolved_dates"]['universe_date']
            except KeyError:
                pass

            factor_ids = None
            try:
                factor_ids = json_openrisk_output['meta']['labels']['factor']['ids']
            except KeyError:
                pass

            if statlevel_level == 'Security':
                if df_security is None:
                    df_security = self.get_security_df(security_ids, universe_date)
                df_security[statlevel_stat] = statlevel_value

            elif statlevel_level == 'Portfolio':
                if df_portfolio is None:
                    df_portfolio = self.get_portfolio_df(universe_date)
                df_portfolio[statlevel_stat] = statlevel_value

            elif statlevel_level == 'Factor':
                if df_factor is None:
                    df_factor = self.get_factor_df(factor_ids, universe_date)
                df_factor[statlevel_stat] = statlevel_value

            elif statlevel_level == 'FactorGroup':
                if df_factorgroup is None:
                    factor_groups = json_openrisk_output['meta']['labels']['factorGroup']['groups']
                    group_names = [group['name'] for group in factor_groups]
                    df_factorgroup = self.get_factor_groups_df(group_names, universe_date, statlevel_stat, statlevel_value)
                else:
                    df_factorgroup[json_openrisk_output['data'][y]['stat']] = statlevel_value

            elif statlevel_level == 'FactorFactor':
                temp_df = self.get_factor_security_df(statlevel_value)

                if temp_df is not None:
                    temp_df.insert(0, "", factor_ids)
                    if df_factorfactor is None:
                        temp_df.insert(0, "date", [universe_date] * len(factor_ids))
                        temp_df.columns = ["date", statlevel_stat] + factor_ids
                        df_factorfactor = temp_df
                    else:
                        temp_df.columns = [statlevel_stat] + factor_ids
                        df_factorfactor = pd.concat([df_factorfactor, temp_df], axis=1)
                        
            elif statlevel_level == 'FactorSecurity':
                temp_df = self.get_factor_security_df(statlevel_value).transpose()
                if temp_df is not None:
                    if df_factorsecurity is None:
                        temp_df.insert(0, "symbol", security_ids)
                        temp_df.insert(0, "date", [universe_date] * len(security_ids))
                        temp_df.columns = ["date", "symbol"] + factor_ids
                        df_factorsecurity = temp_df
                    else:
                        temp_df.insert(0, statlevel_stat, security_ids)
                        temp_df.columns = [statlevel_stat] + factor_ids
                        df_factorsecurity = pd.concat([df_factorsecurity, temp_df], axis=1)
                        
            elif statlevel_level == 'SecuritySecurity':
                temp_df = self.get_factor_security_df(statlevel_value)
                if temp_df is not None:
                    temp_df.insert(0, "", security_ids)
                        
                    if df_securitysecurity is None:
                        temp_df.insert(0, "date", [universe_date] * len(security_ids))
                        temp_df.columns = ["date", statlevel_stat] + security_ids
                        df_securitysecurity = temp_df
                    else:
                        if isinstance(statlevel_value, dict):
                            temp_df.insert(0, "date", [universe_date] * len(security_ids))
                            temp_df.columns = ["date", statlevel_stat] + security_ids
                        else:
                            temp_df.columns = [statlevel_stat] + security_ids
                        df_securitysecurity = pd.concat([df_securitysecurity, temp_df], axis=1)
                        
        # replace None values in dataframe with 0.0
        if (df_portfolio is not None):
            df_portfolio = df_portfolio.fillna(value=0.0)
        if (df_security is not None):
            df_security = df_security.fillna(value=0.0)
        if (df_factor is not None):
            df_factor = df_factor.fillna(value=0.0)
        if (df_factorgroup is not None):
            df_factorgroup = df_factorgroup.fillna(value=0.0)
        if (df_factorfactor is not None):
            df_factorfactor = df_factorfactor.fillna(value=0.0)
        if (df_factorsecurity is not None):
            df_factorsecurity = df_factorsecurity.fillna(value=0.0)
        if (df_securitysecurity is not None):
            df_securitysecurity = df_securitysecurity.fillna(value=0.0)

        return {
            'portfolio': df_portfolio, 
            'security': df_security, 
            'factor': df_factor, 
            'factorGroup': df_factorgroup,
            'factorfactor': df_factorfactor,
            'factorsecurity': df_factorsecurity,
            'securitysecurity': df_securitysecurity
        }

    def generate_matrices_from_df(self, df_raw1, df_raw2):
        """Generates matrices from DataFrames"""
        portfolio_weights = [
            np.matrix(df_raw1['security']["PortfolioWeights"]),
            np.matrix(df_raw2['security']["PortfolioWeights"])
        ]

        benchmark_weights = [
            np.matrix(df_raw1['security']["BenchmarkWeights"]),
            np.matrix(df_raw2['security']["BenchmarkWeights"])
        ]

        residuals = [
            np.matrix(df_raw1['securitysecurity'].drop(['date', 'RawAssetCovarianceMatrix'], axis=1)),
            np.matrix(df_raw2['securitysecurity'].drop(['date', 'RawAssetCovarianceMatrix'], axis=1))
        ]
        
        factor_corr_1 = df_raw1['factorfactor'].set_index('RawFactorCorrelationMatrix').drop(['date'], axis=1)
        factor_corr_2 = df_raw2['factorfactor'].set_index('RawFactorCorrelationMatrix').drop(['date'], axis=1)
        factor_index_union = factor_corr_1.index.union(factor_corr_2.index)
        factor_corr_1 = factor_corr_1.reindex(index=factor_index_union, columns=factor_index_union, fill_value=0.0)
        factor_corr_2 = factor_corr_2.reindex(index=factor_index_union, columns=factor_index_union, fill_value=0.0)
        factor_corr_1 = np.matrix(factor_corr_1)
        factor_corr_2 = np.matrix(factor_corr_2)
        fac_var_1 = df_raw1['factor'].set_index('factor').drop(['date'], axis=1)
        fac_var_2 = df_raw2['factor'].set_index('factor').drop(['date'], axis=1)
        fac_var_1 = fac_var_1.reindex(index=factor_index_union, fill_value=0.0)
        fac_var_2 = fac_var_2.reindex(index=factor_index_union, fill_value=0.0)
        fac_var_1 = np.matrix(fac_var_1)
        fac_var_2 = np.matrix(fac_var_2)

        factor_vols = [
            np.matrix(np.diag(fac_var_1.transpose().tolist()[0])),
            np.matrix(np.diag(fac_var_2.transpose().tolist()[0]))
        ]

        exposure_1 = df_raw1['factorsecurity'].set_index('symbol').drop(['date'], axis=1)
        exposure_2 = df_raw2['factorsecurity'].set_index('symbol').drop(['date'], axis=1)
        exposure_1 = exposure_1.reindex(columns=factor_index_union, fill_value=0.0)
        exposure_2 = exposure_2.reindex(columns=factor_index_union, fill_value=0.0)
        exposure_1 = np.matrix(exposure_1)
        exposure_2 = np.matrix(exposure_2)

        return {
            "portfolio_weights": portfolio_weights, 
            "benchmark_weights": benchmark_weights, 
            "exposures": [ exposure_1, exposure_2 ],
            "factor_corrs": [ factor_corr_1, factor_corr_2],
            "factor_vols": factor_vols, 
            "residuals": residuals, 
        }

    # Total Risk Function - This the manual calc using model inputs, if someone want's to run stats with different formula here we can make the edit.
    def get_total_risk(self, portfolio_weights, benchmark_weights, exposure, factor_corr, fac_var, residual):
        try:
            # initial matrix multiplication
            matmul = exposure @ fac_var @ factor_corr @ fac_var @ exposure.transpose() + residual
            
            # weight difference calculation
            weight_diff = portfolio_weights - benchmark_weights
            
            # active risk calculation
            active_risk = math.sqrt(weight_diff @ matmul @ weight_diff.transpose())
              
            # portfolio risk calculation 
            portfolio_risk = math.sqrt(portfolio_weights @ matmul @ portfolio_weights.transpose())
            
            # benchmark risk calculation
            benchmark_risk = math.sqrt(benchmark_weights @ matmul @ benchmark_weights.transpose())

            return {
                'Active': active_risk, 
                'Portfolio': portfolio_risk,
                'Benchmark': benchmark_risk
            }
        except Exception as e:
            print("Failed to calculate total risk.")
            raise e

    def calculate_delta_risk(self, matrices_dict, date_1, date_2):
        """A function that calculates delta risk"""
        risk_day1 = self.get_total_risk(
            matrices_dict["portfolio_weights"][0], 
            matrices_dict["benchmark_weights"][0],
            matrices_dict["exposures"][0], 
            matrices_dict["factor_corrs"][0], 
            matrices_dict["factor_vols"][0], 
            matrices_dict["residuals"][0]
        )
        self_wb = self.get_total_risk(
            matrices_dict["portfolio_weights"][0], 
            matrices_dict["benchmark_weights"][1],
            matrices_dict["exposures"][0], 
            matrices_dict["factor_corrs"][0], 
            matrices_dict["factor_vols"][0], 
            matrices_dict["residuals"][0]
        )
        self_wp = self.get_total_risk(
            matrices_dict["portfolio_weights"][1], 
            matrices_dict["benchmark_weights"][1],
            matrices_dict["exposures"][0], 
            matrices_dict["factor_corrs"][0], 
            matrices_dict["factor_vols"][0], 
            matrices_dict["residuals"][0]
        )
        self_exposure = self.get_total_risk(
            matrices_dict["portfolio_weights"][1], 
            matrices_dict["benchmark_weights"][1],
            matrices_dict["exposures"][1], 
            matrices_dict["factor_corrs"][0], 
            matrices_dict["factor_vols"][0], 
            matrices_dict["residuals"][0]
        )
        self_factor_volatility = self.get_total_risk(
            matrices_dict["portfolio_weights"][1], 
            matrices_dict["benchmark_weights"][1],
            matrices_dict["exposures"][1], 
            matrices_dict["factor_corrs"][0], 
            matrices_dict["factor_vols"][1], 
            matrices_dict["residuals"][0]
        )
        self_corr = self.get_total_risk(
            matrices_dict["portfolio_weights"][1], 
            matrices_dict["benchmark_weights"][1],
            matrices_dict["exposures"][1], 
            matrices_dict["factor_corrs"][1], 
            matrices_dict["factor_vols"][1], 
            matrices_dict["residuals"][0]
        )
        risk_day2 = self.get_total_risk(
            matrices_dict["portfolio_weights"][1], 
            matrices_dict["benchmark_weights"][1],
            matrices_dict["exposures"][1], 
            matrices_dict["factor_corrs"][1], 
            matrices_dict["factor_vols"][1], 
            matrices_dict["residuals"][1]
        )
        labels = list(risk_day1.keys())
        df_delta = {}
        for label in labels:
            delta_risk_contr = [
                risk_day1[label],
                self_wb[label] - risk_day1[label],
                self_wp[label] - self_wb[label],
                self_exposure[label] - self_wp[label],
                self_factor_volatility[label] - self_exposure[label],
                self_corr[label] - self_factor_volatility[label],  
                risk_day2[label]- self_corr[label],            
                risk_day2[label]
            ]
            delta_level = risk_day2[label] - risk_day1[label]

            try:
                delta_risk_pct_contr = np.array(delta_risk_contr) / delta_level * 100
            except Exception as e:
                print("Failed to calculate delta risk contribution (%).")
                raise e

            delta_risk_pct_contr[[0, -1]] = None
            df_delta[label] = pd.DataFrame(
                {
                    '\u0394 risk Contr.': delta_risk_contr,
                    '%\u0394': delta_risk_pct_contr
                },
                index=[
                    date_1,
                    '\u0394 wB',
                    '\u0394 wP',
                    '\u0394 Exposure',
                    '\u0394 Factor Vol.',
                    '\u0394 Factor Cor.',
                    '\u0394 Residual',
                    date_2
                ]
            )

        # Putting data into df for nicer presentation
        df_delta = pd.concat([df_delta[labels[0]], df_delta[labels[1]], df_delta[labels[2]]], axis=1)
        df_delta.columns = pd.MultiIndex.from_product([labels, ['\u0394 risk Contr.', '%\u0394']])
        return df_delta

    def clean_data(self, port_ids_d1, port_ids_d2, port_mv_d1, port_mv_d2, bench_ids_d1, bench_ids_d2, bench_mv_d1,
                      bench_mv_d2):
        port_data1 = pd.DataFrame({'ids': port_ids_d1, 'port_mv_d1': port_mv_d1})
        port_data2 = pd.DataFrame({'ids': port_ids_d2, 'port_mv_d2': port_mv_d2})
        bench_data1 = pd.DataFrame({'ids': bench_ids_d1, 'bench_mv_d1': bench_mv_d1})
        bench_data2 = pd.DataFrame({'ids': bench_ids_d2, 'bench_mv_d2': bench_mv_d2})
        holdings_df = [port_data1, port_data2, bench_data1, bench_data2]
        holdings_merged = reduce(
            lambda left,
            right: pd.merge(left, right, on='ids', how='outer'),
            holdings_df
        ).fillna(0)

        ids = holdings_merged['ids'].tolist()
        port_mv_1 = holdings_merged['port_mv_d1'].fillna(0).astype(float).tolist()
        port_mv_2 = holdings_merged['port_mv_d2'].fillna(0).astype(float).tolist()
        bench_mv_1 = holdings_merged['bench_mv_d1'].fillna(0).astype(float).tolist()
        bench_mv_2 = holdings_merged['bench_mv_d2'].fillna(0).astype(float).tolist()

        return ids, port_mv_1, port_mv_2, bench_mv_1, bench_mv_2

    def get_delta_output_df(
        self,
        date_1,
        date_2,
        port_ids_d1,
        port_ids_d2,
        port_mv_d1,
        port_mv_d2,
        bench_ids_d1,
        bench_ids_d2,
        bench_mv_d1,
        bench_mv_d2,
        currency,
        calendar,
        risk_model
    ):
        ids, port_mv_d1, port_mv_d2, bench_mv_d1, bench_mv_d2 = \
            self.clean_data(
                port_ids_d1,
                port_ids_d2,
                port_mv_d1,
                port_mv_d2,
                bench_ids_d1,
                bench_ids_d2,
                bench_mv_d1,
                bench_mv_d2
            )
        print("Open:Risk API: Calculating")
        response_1 = self.calculate_open_risk_holdings(
            date_1,
            ids,
            port_mv_d1,
            ids,
            bench_mv_d1,
            currency,
            calendar,
            risk_model
        )
        response_2 = self.calculate_open_risk_holdings(
            date_2,
            ids,
            port_mv_d2,
            ids,
            bench_mv_d2,
            currency,
            calendar,
            risk_model
        )
        print("Open:Risk API: Calculation route: Complete")
        df_raw1 = self.convert_json_output_to_df(response_1)
        df_raw2 = self.convert_json_output_to_df(response_2)
        matrices_dict = self.generate_matrices_from_df(df_raw1, df_raw2)
        final_output = self.calculate_delta_risk(matrices_dict, date_1, date_2)
        return final_output


class HoldingsReader:
    def __init__(self):
        pass

    def read_csv(self, file_name):
        try:
            df = pd.read_csv(file_name)
        except FileNotFoundError:
            print(f"{file_name} not found.")
            return []
        if 'ids' not in df.columns or 'mv' not in df.columns:
            print(f"Required columns not found in {file_name}")
            return []

        return df['ids'].tolist(), df['mv'].tolist()

    def load_data(self, port1, port2, bench1, bench2):
        port_ids_1, port_mv_1 = self.read_csv(port1)
        port_ids_2, port_mv_2 = self.read_csv(port2)
        bench_ids_1, bench_mv_1 = self.read_csv(bench1)
        bench_ids_2, bench_mv_2 = self.read_csv(bench2)

        return port_ids_1, port_ids_2, port_mv_1, port_mv_2, bench_ids_1, bench_ids_2, bench_mv_1, bench_mv_2


